@@ -51,6 +51,11 @@ SIMPLESE_VERSION = re.compile('MPU_A4D[A-Za-z\t .]+')
 '''
 ALL_OFF_PATTERN = re.compile('ALLOFF')
 '''
+#ZxxLKlll
+'''
+KEYPAD_LOCK_PATTERN = re.compile('Z(?P<zone>\d{2})'
+                     'LK(?P<lock>ON|OFF)')
+'''
 #?
 '''
 ERROR_PATTERN = re.compile('\?')
@@ -59,6 +64,7 @@ ERROR_PATTERN = re.compile('\?')
 EOL = b'\r'
 TIMEOUT_OP       = 0.2   # Number of seconds before serial operation timeout
 TIMEOUT_RESPONSE = 1     # Number of seconds before command response timeout
+page_active = np.array([0,0])
 zoneinit = np.array([1,0,0,0,0,0,0,0,0,0,0,0,0])
 zonesetinit = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonepwr = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
@@ -70,6 +76,7 @@ zonesrc = np.array([1,0,0,0,0,0,0,0,0,0,0,0,0])
 zonesrcsave = np.array([0,1,1,1,1,1,1,1,1,1,1,1,1])
 zonemute = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonemutesave = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
+zonelock = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonebass = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonetreble = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0])
 zonegroupsrc = np.array([1,1,1,1,1,1,1,1,1,1,1,1,1])
@@ -179,6 +186,10 @@ class ZonesetStatus(object):
            self.group = bool(0)
         else:
            self.group = bool(1)
+        if zonelock[self.zone] == 0:
+           self.keypad_lock = bool(0)
+        else:
+           self.keypad_lock = bool(1)
 
         if from_port:
             if 'settings_zonetable' in globals():
@@ -261,6 +272,14 @@ class Nuvo(object):
         """
         raise NotImplemented()
 
+    def set_keypad_lock(self, zone: int, lock: bool):
+        """
+        Lock or unlock zone keypad
+        :param zone: zone 1.12
+        :param mute: True to lock, False to unlock
+        """
+        raise NotImplemented()
+
     def set_volume(self, zone: int, volume: float):
         """
         Set volume for zone
@@ -380,6 +399,12 @@ def _parse_response(string: bytes):
           _LOGGER.debug('ALL OFF received')
           all_off_keypad()
 
+   if not match:
+       match = re.search(KEYPAD_LOCK_PATTERN, string)
+       if match:
+          _LOGGER.debug('Keypad lock status received')
+          keypad_lock_recv(match)
+
    if not match:
        match = re.search(ERROR_PATTERN, string)
        if match:
@@ -433,6 +458,14 @@ def _format_set_mute(zone: int, mute: bool) -> str:
        zonemute[int(zone)] = 0
        return 'Z{:0=2}MTOFF'.format(int(zone))
 
+def _format_set_keypad_lock(zone: int, lock: bool) -> str:
+    if (lock):
+       zonelock[int(zone)] = 1
+       return 'Z{:0=2}LKON'.format(int(zone))
+    else:
+       zonelock[int(zone)] = 0
+       return 'Z{:0=2}LKOFF'.format(int(zone))
+
 def _format_set_group(zone: int, group: bool) -> str:
     if (group):
        return 'Z{:0=2}GRPON'.format(int(zone))
@@ -484,20 +517,29 @@ def _format_all_off() -> str:
 def all_off_keypad():
     time.sleep(0.5) # Nuvo does NOT process commands immediately after ALLOFF
     zonesalloff = 1
-    for count in range(1,13):
+    for count in range(0,13):
         if zonepwr[int(count)] == 1:
             zonesalloff = 0
-    for count in range(1,13):
+    for count in range(0,13):
         if zonesalloff == 0:
             zonepwralloff[int(count)] = zonepwr[int(count)]
             zonepwr[int(count)] = 0
         else:
             if zonepwralloff[int(count)] == 1 and ALL_OFF_RECALL:
                 zonepwr[int(count)] = 1
-        if count in media_zonetable:
-            callbackobj = media_zonetable.index(count)
-            callbackint = int(callbackobj)
-            media_callback[callbackint]()
+
+def keypad_lock_recv(match):
+    zone = abs(int(match[1]))
+    lock = match[2]
+    if lock == 'ON':
+        zonelock[zone] = 1
+    else:
+        zonelock[zone] = 0
+    if zone in settings_zonetable:
+        callbackindex = [i for i, x in enumerate(settings_zonetable) if x == zone]
+        for callbackzone in callbackindex:
+            settings_callback[callbackzone]()
+
 
 def get_nuvo(port_url, baud, all_off_recall):
     """
@@ -791,6 +833,17 @@ def get_nuvo(port_url, baud, all_off_recall):
             else:
                 self._process_request(_format_set_mute(zone, mute))
 
+        @synchronized
+        def set_keypad_lock(self, zone: int, lock: bool):
+            if zone == 0:
+                if lock:
+                    zonelock[0] = 1
+                else:
+                    zonelock[0] = 0
+                media_callback[0]()
+            else:
+                self._process_request(_format_set_keypad_lock(zone, lock))
+
         @synchronized
         def set_volume(self, zone: int, volume: float):
             if zone == 0:
@@ -833,6 +886,9 @@ def get_nuvo(port_url, baud, all_off_recall):
 
         @synchronized
         def page_off(self, page_source, page_zones):
+            while page_active[1] == 1:
+                _LOGGER.info('Paging on service call currently in progress!  Waiting to turn off paging.')
+                time.sleep(1)
             zonecnt = 0
             while zonecnt < len(page_zones):
                self.set_volume(int(page_zones[zonecnt]),
@@ -846,24 +902,31 @@ def get_nuvo(port_url, baud, all_off_recall):
                        self.set_source(int(page_zones[zonecnt]),
                                        zonesrcsave[int(page_zones[zonecnt])])
                zonecnt += 1
+            page_active[0] = 0
 
         @synchronized
         def page_on(self, page_source, page_zones, page_volume):
-            zonecnt = 0
-            while zonecnt < len(page_zones):
-               _LOGGER.debug('Page Zone: %s, Vol: %s, Src: %s', page_zones[zonecnt],\
-                   page_volume[zonecnt], page_source)
-               zonepwrsave[int(page_zones[zonecnt])] = zonepwr[int(page_zones[zonecnt])]
-               zonesrcsave[int(page_zones[zonecnt])] = zonesrc[int(page_zones[zonecnt])]
-               zonevolsave[int(page_zones[zonecnt])] = zonevol[int(page_zones[zonecnt])]
-               zonemutesave[int(page_zones[zonecnt])] = zonemute[int(page_zones[zonecnt])]
-               if zonepwr[int(page_zones[zonecnt])] == 0:
-                   self.set_power(int(page_zones[zonecnt]), True)
-                   time.sleep(0.05)
-               if zonesrc[int(page_zones[zonecnt])] != page_source:
-                   self.set_source(int(page_zones[zonecnt]), page_source)
-               self.set_volume(int(page_zones[zonecnt]), percent_to_db(int(page_volume[zonecnt])))
-               zonecnt += 1
+            if page_active[0] == 0:
+                page_active[1] = 1
+                zonecnt = 0
+                while zonecnt < len(page_zones):
+                   _LOGGER.debug('Page Zone: %s, Vol: %s, Src: %s', page_zones[zonecnt],\
+                       page_volume[zonecnt], page_source)
+                   zonepwrsave[int(page_zones[zonecnt])] = zonepwr[int(page_zones[zonecnt])]
+                   zonesrcsave[int(page_zones[zonecnt])] = zonesrc[int(page_zones[zonecnt])]
+                   zonevolsave[int(page_zones[zonecnt])] = zonevol[int(page_zones[zonecnt])]
+                   zonemutesave[int(page_zones[zonecnt])] = zonemute[int(page_zones[zonecnt])]
+                   if zonepwr[int(page_zones[zonecnt])] == 0:
+                       self.set_power(int(page_zones[zonecnt]), True)
+                       time.sleep(0.05)
+                   if zonesrc[int(page_zones[zonecnt])] != page_source:
+                       self.set_source(int(page_zones[zonecnt]), page_source)
+                   self.set_volume(int(page_zones[zonecnt]), percent_to_db(int(page_volume[zonecnt])))
+                   zonecnt += 1
+                   page_active[0] = 1
+                page_active[1] = 0
+            else:
+                _LOGGER.info('Paging already active, ignoring second request.')
 
         @synchronized
         def mute_all(self):
@@ -888,7 +951,7 @@ def get_nuvo(port_url, baud, all_off_recall):
         @synchronized
         def all_off(self):
             self._process_request(_format_all_off())
-            for count in range(1,13):
+            for count in range(0,13):
                 zonepwralloff[int(count)] = 0
                 zonepwr[int(count)] = 0
                 if count in media_zonetable:
